/*
 * Copyright (C) 2013-2015 by the University of Southern California
 * $Id: parse.cc 24683 2015-11-25 07:29:11Z johnh $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <signal.h>
#include <netinet/in.h>
#include <climits>
#include <iostream>
#include <iomanip>
#include <string>
#include <ldns/ldns.h>

#include "parse.hh"
#include "table.hh"
#include "libtrace.h"
#include "scramble_crypt.h"
#include "global_var.h"
#include "anon.hh"
#include "utility.hh"

#include "dsc_glue.hh"

#define timer_assign(a, b)						\
	do {								\
		(a)->tv_sec = (b)->tv_sec;				\
		(a)->tv_usec = (b)->tv_usec;				\
	} while (0)

//static const char *SERVER_ID_FNAME = "global_server_id.fsdb";
static const std::string PAYLOADZERO = "load0";
static const std::string ERRORCODE = "err";

using namespace std;

/*
	define the schema; this kind of vector initialization is valid in c++11: -std=c++11
*/
vector<string> PT_sch {	//packet table
	"msgid",	//unique message id, packet identifier
	"time",		//time stamp 
	"srcip",	//src ip address
	"srcport",	//src port number
	"dstip",	//dst ip
	"dstport",	//dst port number
	"protocol",	//TCP or UDP
	"id",           // dns query/reply id 
	"qr", "opcode", "aa", "tc", "rd", "ra", "z", "ad", "cd", "rcode",	//DNS flag
	"qdcount", "ancount", "nscount", "arcount", 	//RR counters, note the OPT (EDNS)
							//in addtional field does not count
	"edns_present", "edns_udp_size", "edns_extended_rcode", "edns_version", "edns_z", //EDNS
	"msglen"
};

vector<string> non_unified_sch {
	"msgid",	//unique message id, packet identifier
	"id"           // dns query/reply id
};

vector<string> partial_QT_sch {	//question table
	"name",		//question name
	"type",		//question type
	"class"		//question class                             
};

vector<string> partial_RT_sch{	//rr table
	"section_name",	//in which section of the packet. answer(ANS), authority(AUT), additional(ADD)
	"name", "type", "class",	// answer, authority, additional
	"ttl",
	"rdf_num", //number of fields in rdata
	"rdata"    //rdata string, spaces are mapped to 20%
};

vector<string> TCPET_sch{	//error parsing pkt table
	"time",		//time stamp
	"srcip",	//source ip, we cannot define which is user, just log src and dst
	"srcport",	//source port
	"dstip",	//destination ip
	"dstport",	//destination port
	"payload_size",	//packet payload size
	"state",	//PAYLOADZERO: payload is 0
			//ERRORCODEX: error code is X
	"ecn_ns",	//ECN Nonce Sum.
	"doff",		//Data Offset.
	"fin",		//FIN
	"syn",		//syn
	"rst",		//RST flag.
	"psh",		//PuSH flag.
	"ack",		//ACK flag.
	"urg",		//URG flag.
	"ece",		//ECN Echo.
	"cwr",		//ECN CWR. 
	"window"	//window size
};

vector<string> UDPET_sch  {	//error parsing pkt table
	"time",		//time stamp
	"srcip",	//source ip, we cannot define which is user, just log src and dst
	"srcport",	//source port
	"dstip",	//destination ip
	"dstport",	//destination port
	"payload_size",	//packet payload size
	"state"		//PAYLOADZERO: payload is 0
			//ERRORCODEX: error code is X 
};

/* 
 * rssac022_sch should match the output of message_to_rssacint.
 * See that program for definitions.
 */
vector<string> RSSAC002_sch {
	"key",
	"count"
};

///packet counter, just line number
long long msgCount = 0;
bool firstpkt = true;	
string output_prefix;
vector<string> compressor;
long output_count = 0;

//create the table, they will be set when dealing with the first packt 
Table *PT = NULL;
Table *QT = NULL;
Table *RT = NULL;
Table *TCPET = NULL;
Table *UDPET = NULL;
Table *RSSAC002T = NULL;

//the key and initialization vector; used for ip anonymization and domain name encryption
unsigned char *enc_key = NULL;
unsigned char *enc_iv = NULL;
int key_len = 0;
int iv_len = 0;

//time to clean up tcp
struct timeval check_tcp_ts;

/*
	clean a single table
*/
void clean_table (Table *t) 
{
	if (t != NULL) {
		// print a comment about how we were run
		t->final(command_line);
		delete t;
	}
}

/*
	clean all the tables
*/
void clean_all_table () 
{
	if (verbose_log)
		cout<<"clean all the tables"<<endl;
		
	clean_table(PT);
	clean_table(QT);
	clean_table(RT);
	clean_table(TCPET);
	clean_table(UDPET);
	clean_table(RSSAC002T);
}

/*
	signal handler
*/
void sig_handler (int sig)
{
	clean_all_table();
	cout<<"user kills the program."<<endl;
	exit(0);
}

/*
	libtrace clean up function copied from examples of libtrace
	standard libtrace code
*/
static void libtrace_cleanup(libtrace_t *trace, libtrace_packet_t *packet)
{
	if (trace) {
		trace_destroy(trace);
	}
		
	if (packet) {
		trace_destroy_packet(packet);
	}
}

/*
	format rdata string, replace the spaces by "%20"
*/
void format_rdata(string &s)
{	
	istringstream ss(s);
	s = "";
	while (!ss.eof()) {	//spaces, tab will be replaced by "%20"
		string x;
		ss >> x;
		s += x;
		if (!ss.eof()) {
			s += RDATA_SPT;
		}
	}
}

/*
	process each section, QUE, AUS, AUT, ADD
*/
void
do_section(ldns_pkt *m, ldns_pkt_section sec, int msg_ct, string dns_id, Table *t, bool is_anon,
    unsigned char *key, size_t len, unsigned char *iv, size_t iv_len, int pbits,
    string prefix_str)
{
	stringstream ss;
	if (!t) return;
	
	ldns_rr_list *rrs = NULL;
	//ldns_pkt_section_count exists but is not decleared in ldns-1.6.16
	//uint16_t rr_count = ldns_pkt_section_count(m, sec);
	uint16_t rr_count = 0;
	
	string type;
	switch (sec) {
	case LDNS_SECTION_QUESTION:
		type = "QUE";
		rrs = ldns_pkt_question(m);
		rr_count = ldns_pkt_qdcount(m);
		break;		
	case LDNS_SECTION_ANSWER:
		type = "ANS";
		rrs = ldns_pkt_answer(m);
		rr_count = ldns_pkt_ancount(m);
		break;
	case LDNS_SECTION_AUTHORITY:
		type = "AUT";
		rrs = ldns_pkt_authority(m);
		rr_count = ldns_pkt_nscount(m);
		break;				
	case LDNS_SECTION_ADDITIONAL:
		type = "ADD";
		rrs = ldns_pkt_additional(m);
		rr_count = ldns_pkt_arcount(m);
		break;
	default:
		type = "unknown";
		rr_count = 0;
	}		

	if (!rrs) {
		warnx("[warn] unknown section");
		return;
	}

	//handle empty questions for unified output -p Q:
	//no question content but we still need to log the message
	if (report_qry == UNIFIED_REPORT && show_empty_query &&
	    !prefix_str.empty() && sec == LDNS_SECTION_QUESTION && rr_count == 0) {
		ss.str("");
		ss << prefix_str << "\t-\t-\t-" << endl;
		t->put(ss.str());
	}

	for (unsigned int i=0; i < rr_count; i++) {
		ldns_rr *rr = ldns_rr_list_rr(rrs, i);
		char *rr_class = ldns_rr_class2str(ldns_rr_get_class(rr));

		//ldns (1.6.17) cannot print TYPE AXFR; TYPE252 instread?
		//ldns (1.6.17) cannot print TYPE ANY; TYPE255 instread?
		char *rr_type = ldns_rr_type2str(ldns_rr_get_type(rr));
		if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_AXFR) {
			LDNS_FREE(rr_type);
			rr_type = (char *)"AXFR";
		}
		if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_ANY) {
			LDNS_FREE(rr_type);
			rr_type = (char *)"ANY";
		}
		char *name = ldns_rdf2str(ldns_rr_owner(rr));
		string hashname = name;
		
		if (is_anon)				//anonymize the name of the RR
			anon_name(name, key, len, iv, iv_len, hashname, false);			
		
		if (sec == LDNS_SECTION_QUESTION) {	//questions			
			ss.str("");
			if (prefix_str.empty()) {
				ss << msg_ct << "\t" << dns_id;
			} else { // unified output, so grab the old line
				ss << prefix_str;
			}
			ss << "\t" << hashname << "\t" 
			   << rr_type << "\t" << rr_class << endl;
			string line = ss.str(); 
			t->put(line);		
		} else {				//others: answer, authority, addtional
			//deal with rdata first
			string rdata_str;
			/*
			  time in rdata will be printed as UTC time string like "YYYYMMDDHHMMSS"
			  from ldns (1.6.17) code:
			  host2str.c: ldns_rdf2buffer_str() -> ldns_rdf2buffer_str_fmt() ->
			  LDNS_RDF_TYPE_TIME -> ldns_rdf2buffer_str_time ->
			  util.c: ldns_serial_arithmitics_gmtime_r
			 */
			for (unsigned int j = 0; j<ldns_rr_rd_count(rr); j++) {
				if (ldns_rr_rdf(rr, j)) {
					char *tmp = ldns_rdf2str(ldns_rr_rdf(rr, j));
					if (!rdata_str.empty()) rdata_str += " ";
					rdata_str += tmp;
					LDNS_FREE(tmp);
				}
			}
			
			if (rdata_str.empty()) {
				rdata_str = "-";
			} else {
				rdata_str.erase(rdata_str.find_last_not_of(" \t") + 1);
				rdata_str.erase(0, rdata_str.find_first_not_of(" \t"));
				if (is_anon) {		
					anon_rdata(ldns_rr_get_type(rr), rdata_str, key, len, iv, iv_len, pbits);
				} else {
					format_rdata(rdata_str); //get rid of spaces and tab
				}	
			}
				
			//output to file
			ss.str("");
			if (prefix_str.empty()) {
				ss << msg_ct << "\t" << dns_id;
			} else {
				// unified output, so grab the old line
				ss << prefix_str;
			};
			ss << "\t" << type << "\t" << hashname << "\t"
			   << rr_type << "\t" << rr_class << "\t"
			   << ldns_rr_ttl(rr) << "\t" << ldns_rr_rd_count(rr) << "\t"
			   << rdata_str << endl;
			string line = ss.str();
			t->put(line);		
		}
		LDNS_FREE(rr_class);
		if (ldns_rr_get_type(rr) != LDNS_RR_TYPE_AXFR && ldns_rr_get_type(rr) != LDNS_RR_TYPE_ANY)
			LDNS_FREE(rr_type);
		LDNS_FREE(name);
	}
}

/*
	out_extra_info
*/
void out_extra_info(const string &s1, string &state, libtrace_tcp_t *tcp_header, bool istcp, Table *t)
{
	stringstream ss;
	if (istcp) {
		ss << tcp_header->ecn_ns << "\t" << tcp_header->doff << "\t" 
		   << tcp_header->fin << "\t" << tcp_header->syn << "\t" 
		   << tcp_header->rst << "\t" << tcp_header->psh << "\t" 
		   << tcp_header->ack << "\t" << tcp_header->urg << "\t" 
		   << tcp_header->ece << "\t" << tcp_header->cwr << "\t" 
		   << tcp_header->window;			
		string s2 = ss.str();
		ss.str("");
		ss << s1 << "\t" << state << "\t" << s2 << endl;
		string err_string = ss.str();
		t->put(err_string);
	} else {//udp
		ss << s1 << "\t" << state << "\t" << endl;
		string err_string = ss.str();			
		t->put(err_string);
	}
}

/*
	get the enc key from the given keyfile
*/
void get_enc_key (string &keyFile)
{
	scramble_state_t scr_st;	//for getting the key 
	const char *keyFileName = (const char *)keyFile.c_str();
	if (scramble_init_from_file(keyFileName, SCRAMBLE_BLOWFISH, SCRAMBLE_BLOWFISH, NULL) < 0) {
		cerr << "error reading keyfile " << keyFileName << endl;
		exit(-1);
	}
	
	//get the key for hashing
	const int max_blk_length = 32;//scramble_crypt.cc: define MAX_BLK_LENGTH	32
	
	/*char ppad[max_blk_length];
	char kkey[max_blk_length];
	char mmac[max_blk_length];
	char iiv[max_blk_length];*/
	u_char ppad[max_blk_length];
	u_char kkey[max_blk_length];
	u_char mmac[max_blk_length];
	u_char iiv[max_blk_length];	
	scr_st.iv	= iiv;
	scr_st.ivlen	= max_blk_length;
	
	scr_st.pad 	= ppad;	
	scr_st.key 	= kkey;	
	scr_st.mac 	= mmac;	
	scr_st.plen 	= max_blk_length;
	scr_st.klen 	= max_blk_length;
	scr_st.mlen 	= max_blk_length;
	if (scramble_readstate(keyFileName, &scr_st) < 0) {
		cerr<<"[ERROR] scramble_readstate"<<endl;
		exit(-1);
	}		
	key_len = scr_st.klen; 
	enc_key = new unsigned char[key_len];
	memcpy(enc_key, scr_st.key, key_len);	
		
	iv_len = scr_st.ivlen;
	enc_iv = new unsigned char[iv_len];
	memcpy(enc_iv, scr_st.iv, iv_len);
#ifdef DEBUG	
	cout << "key file name is: " << keyFileName << endl
	<< "key len =" << scr_st.klen << endl;
	cout << "pad len =" << scr_st.plen << endl;
	cout << "mac len =" << scr_st.mlen << endl;	
	cout << "iv len =" << scr_st.ivlen << endl;
#endif
}

map<string, string> compressor_suffix_map;

string
determine_compressor_suffix(string compressor)
{
	string prog;
	size_t slash_i = compressor.rfind('/');
	if (slash_i != string::npos) {
		prog = compressor.substr(slash_i + 1);
	} else {
		prog = compressor;
	};
	string suffix = "";

	if (compressor_suffix_map.empty()) {
		compressor_suffix_map["xz"] = "xz";
		compressor_suffix_map["gzip"] = "gz";
		compressor_suffix_map["bzip2"] = "bz2";
		compressor_suffix_map["none"] = "";
		compressor_suffix_map["unknown"] = "ZZ";
	};
	map<string,string>::iterator it = compressor_suffix_map.find(prog);
	if (it == compressor_suffix_map.end()) // a default!
		it = compressor_suffix_map.find("unknown");
	if (it == compressor_suffix_map.end()) {
		cerr << "dnsanon: cannot determine suffix for compressor " + compressor + " (please improve parse.cc:determine_compressor_suffix)\n";
		exit(1);
	};
	return it->second;
}

/*
  get the ipv4 ttl or ipv6 hop limit
*/

uint8_t get_ip_ttl(libtrace_packet_t *packet, bool &err)
{
	uint16_t ethertype = 0;
	uint32_t remaining = 0;
	uint8_t ttl = 0;
	void *ip_hdr = trace_get_layer3(packet, &ethertype, &remaining);
	err = false;
	if (ip_hdr == NULL) {
		err = true;
		return ttl;
	}
	if (ethertype == TRACE_ETHERTYPE_IP) {
		libtrace_ip_t *ip = (libtrace_ip_t *)ip_hdr;
		ttl = ip->ip_ttl;
	} else if (ethertype == TRACE_ETHERTYPE_IPV6) {
		libtrace_ip6_t *ip6 = (libtrace_ip6_t *)ip_hdr;
		ttl = ip6->hlim;
	} else {
		err = true;
	}
	return ttl;
}

/*
	deal with each message

	(post TCP or UDP)

*/
void
per_message(libtrace_packet_t *packet, u_char *payload, int payload_len, string &outputPath, int pbits)
{
	//deal with msg counter overflow
	if (msgCount >= LLONG_MAX-100) {		//start a new set of files
		/* xxx: hard to believe this path has been tested */
		msgCount = 0;
		firstpkt = true;
		clean_all_table();		//close all the previous tables
	}

	uint32_t remaining = 0;
	uint8_t proto = 0;
	string protocol_str = "unknown";
	void *transport_header = trace_get_transport(packet, &proto, &remaining);	

	if (proto == IPPROTO_UDP) {
		protocol_str = "udp";
	} else if (proto == IPPROTO_TCP) {
		protocol_str = "tcp";
	} else {
		/* xxx: should log it */
		return;  // unknown protocol
	};

	uint16_t src_port = 0;			//get port numbers
	uint16_t dst_port = 0;
	dst_port = trace_get_destination_port(packet);
	src_port = trace_get_source_port(packet); 
	
	if (dst_port != 53 && src_port != 53) {	//not DNS packet
		return;
	}

	//get ip ttl
	bool ip_ttl_err = false;
	uint8_t ip_ttl = get_ip_ttl(packet, ip_ttl_err);
	string ip_ttl_str = ip_ttl_err ? "-":to_string(ip_ttl);
	
	// use this stringstream for generic formatting
	stringstream ss;
	
	// time stamp
	struct timeval msg_ts = trace_get_timeval(packet);
	ss  << msg_ts.tv_sec << "." << setfill('0') << setw(6) << msg_ts.tv_usec;
	string timestamp_str = ss.str();

	// clean up old tcp
	if (tcp_timeout > 0) {
		bool new_check_tcp_ts = false;
		int next_check_interval = tcp_check_period;
		if (!timerisset(&check_tcp_ts)) {//this is the first packet/message
			new_check_tcp_ts = true;
			if (tcp_timeout > tcp_check_period) {
				//no tcp-timeout within the default next_check_interval (tcp_check_period)
				//after the first clean up, next_check_interval will be set to tcp_check_period
				next_check_interval = tcp_timeout;
				//debug: cerr << "set next_check_interval: " << tcp_timeout << endl;
			}
		} else if (msg_ts.tv_sec > check_tcp_ts.tv_sec) { //time to clean up old tcp connection
			new_check_tcp_ts = true;
			dnsanon_tcpList_remove_older_than(msg_ts.tv_sec - tcp_timeout);
			//debug: cerr << "dnsanon_tcpList_remove_older_than @ " << timestamp_str << endl;
		}
		if (new_check_tcp_ts) {
			timer_assign(&check_tcp_ts, &msg_ts);
			check_tcp_ts.tv_sec += next_check_interval;
			//debug: cerr << "next check @ " << check_tcp_ts.tv_sec << endl;
		}
	}
		
	//we need the timestamp of the first packet as the output prefix
	if (firstpkt) {
		string file_prefix = outputPath;
		if (file_prefix.back() != '/')
			file_prefix += "/";
		string file_suffix = ".fsdb";
		string compressor_suffix = determine_compressor_suffix(compressor[0]);
		if (!compressor_suffix.empty()) {
			file_suffix += "." + compressor_suffix;
		};
		if (output_prefix.length() == 0)
			file_prefix += timestamp_str;
		else {
			file_prefix += output_prefix;
			if (output_count != 0)
				file_prefix += ("." + to_string(output_count));
			output_count += 1;
		}
		if (report_messages) {
			PT = new Table(PT_sch, (file_prefix + ".message" + file_suffix), compressor);
		};
		if (report_qry != NO_REPORT) {
			vector<string> QT_sch;
			if (report_qry == UNIFIED_REPORT)
				QT_sch.insert(QT_sch.end(), PT_sch.begin(), PT_sch.end());
			else QT_sch.insert(QT_sch.end(), non_unified_sch.begin(), non_unified_sch.end());
			QT_sch.insert(QT_sch.end(), partial_QT_sch.begin(), partial_QT_sch.end());
			QT = new Table(QT_sch, (file_prefix + (report_qry == UNIFIED_REPORT ? ".message_question" : ".question") + file_suffix), compressor);
		};
		if (report_rsp != NO_REPORT) {
			vector<string> RT_sch;
			if (report_rsp == UNIFIED_REPORT)
				RT_sch.insert(RT_sch.end(), PT_sch.begin(), PT_sch.end());
			else RT_sch.insert(RT_sch.end(), non_unified_sch.begin(), non_unified_sch.end());
			RT_sch.insert(RT_sch.end(), partial_RT_sch.begin(), partial_RT_sch.end());
			RT = new Table(RT_sch, (file_prefix + (report_rsp == UNIFIED_REPORT ? ".message_rr" : ".rr") + file_suffix), compressor);
		};
		if (report_extra) {
			TCPET = new Table(TCPET_sch, (file_prefix + ".tcp_extra" + file_suffix), compressor);
			UDPET = new Table(UDPET_sch, (file_prefix + ".udp_extra" + file_suffix), compressor);
		};
		if (report_rssac002 != NO_REPORT) {
			RSSAC002T = new Table(RSSAC002_sch, (file_prefix+".rssacint"+file_suffix), compressor);
		};
		firstpkt = false;	
	}
	
	//get ip addresses	
	struct sockaddr_storage src_addr;
	struct sockaddr_storage dst_addr;
	struct sockaddr *src_ip = NULL;
	struct sockaddr *dst_ip = NULL;
	src_ip = trace_get_source_address(packet, (struct sockaddr *)&src_addr);
	dst_ip = trace_get_destination_address(packet, (struct sockaddr *)&dst_addr);
	
	//clear addresses, we might need clear server address for monitoring
	string clear_src_ip_str;
	string clear_dst_ip_str;
	get_ip_str(src_ip, clear_src_ip_str, false, -1);
	get_ip_str(dst_ip, clear_src_ip_str, false, -1);		
	
	//prefix-preserving ip addresses
	string anon_src_ip_str;
	string anon_dst_ip_str;

	bool do_anon_src = (
			    (src_port == 53 && ((anon_flag & ANON_SERVER_ADDR) != ANON_NONE)) ||
			    (src_port != 53 && ((anon_flag & ANON_CLIENT_ADDR) != ANON_NONE))
			    );
	bool do_anon_dst = (
			    (dst_port == 53 && ((anon_flag & ANON_SERVER_ADDR) != ANON_NONE)) ||
			    (dst_port != 53 && ((anon_flag & ANON_CLIENT_ADDR) != ANON_NONE))
			    );
	
	get_ip_str(src_ip, anon_src_ip_str, do_anon_src, pbits);
	get_ip_str(dst_ip, anon_dst_ip_str, do_anon_dst, pbits);
			

#ifdef DEBUG		
	//cout << "time=" << timestamp << endl;
	//printf("time=%u", ts.tv_sec);	
	//cout<<protocol<<endl;
	cout << "msgCount=" << msgCount << endl;
	cout << "remaining="<<remaining<< endl;//continue;
#endif
	ldns_pkt *m = NULL;
	ldns_status s = ldns_wire2pkt(&m, payload, payload_len);
	if(s != LDNS_STATUS_OK) {
		ldns_pkt_free(m);
		if (verbose_log)
			warnx("cannot parse dns packet:%s", ldns_get_errorstr_by_id(s));
		if (report_extra) {
			// xxx: need to merge error handling for tcp and and udp, and above
			//error in parsing, log basic pkt info, go to next one
			ss.str("");
			ss << ERRORCODE << (int)s;
			string err_str = ss.str();
			ss.str("");
			ss << timestamp_str << "\t" << anon_src_ip_str << "\t" << src_port << "\t"
			   << anon_dst_ip_str << "\t" << dst_port << "\t" << remaining;
			if (proto == IPPROTO_TCP) {
				out_extra_info(ss.str(), err_str, (libtrace_tcp_t *)transport_header, true, TCPET);
			} else if (proto == IPPROTO_UDP) {
				out_extra_info(ss.str(), err_str, NULL, false, UDPET);
			}
		}
		return;
	}
	if (!m) {
		warn("memory problem, parsed message is NULL");
		return;
	}

	/*bool is_rsp = ldns_pkt_qr(m);
	if (!PT && ((is_rsp && report_rsp == NO_REPORT) || (!is_rsp && report_qry == NO_REPORT)) ){
		// we don't care about this message, so move on
		ldns_pkt_free(m);
		return;
	}*/
	
	//every thing is good if getting here
	msgCount++;
		
	ss.str("");
	ss << ldns_pkt_id(m);
	string dnsid_str = ss.str();
	
	if ((anon_flag & ANON_DNS_ID) != ANON_NONE) {	//anonymize id
		string anon_dnsid_str;
		anon_string((char *)dnsid_str.c_str(), enc_key, key_len, 
			enc_iv, iv_len, anon_dnsid_str, false);	
		dnsid_str = anon_dnsid_str;
	}

	/*
	//http://www.iana.org/assignments/dns-parameters/dns-parameters.xml#dns-parameters-11
	if (full_anonymization)
		string edns_options = "-1";//not sure if it's anonymized or not, leave it empty here
	*/

	/*
	  ldns: arcount does not include EDNS and TSIG, see wire2host.c:
	  if(have_edns)
	    ldns_pkt_set_arcount(packet, ldns_pkt_arcount(packet) - have_edns);

          if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_TSIG) {
	    ldns_pkt_set_tsig(packet, rr);
	    ldns_pkt_set_arcount(packet, ldns_pkt_arcount(packet) - 1);
	  }
	  xxx: to make output consistent, let's add tsig rr back to additional section
	 */

	if (ldns_pkt_tsig(m)) {
		if (!ldns_rr_list_push_rr(ldns_pkt_additional(m), ldns_rr_clone(ldns_pkt_tsig(m)))) {
			ldns_pkt_free(m);
			return;
		}
		ldns_pkt_set_arcount(m, ldns_pkt_arcount(m) + 1);
	}
	
	//if pass (int8_t)ldns_pkt_edns_version(m) directly to stringstream, 
	//the rest of the stringstream will be ignored
	//the first two bytes are 0 which is take as '\0' end of string?
	//the same for ends_extended_rcode
	int edns_version = ldns_pkt_edns_version(m);
	int ends_extended_rcode = ldns_pkt_edns_extended_rcode(m);

	// ldns does not have a API of "ldns_pkt_z()"
	// but it has Macros: LDNS_Z_WIRE, see wire2host.h
	ss.str("");
	if (PT || QT || RT) {
		ss << msgCount << "\t" << timestamp_str << "\t"
		   << anon_src_ip_str << "\t" << src_port << "\t"
		   << anon_dst_ip_str << "\t" << dst_port << "\t"
		   << protocol_str << "\t" << dnsid_str << "\t" 
		   << ldns_pkt_qr(m) << "\t" << ldns_pkt_get_opcode(m) << "\t" 
		   << ldns_pkt_aa(m) << "\t" << ldns_pkt_tc(m) << "\t" 
		   << ldns_pkt_rd(m) << "\t" << ldns_pkt_ra(m) << "\t" 
		   << LDNS_Z_WIRE(payload) << "\t"
		   << ldns_pkt_ad(m) << "\t" << ldns_pkt_cd(m) << "\t"
		   << ldns_pkt_get_rcode(m) << "\t"
		   << ldns_pkt_qdcount(m) << "\t" << ldns_pkt_ancount(m) << "\t" 
		   << ldns_pkt_nscount(m) << "\t" << ldns_pkt_arcount(m) << "\t" 
		   << ldns_pkt_edns(m) << "\t" << ldns_pkt_edns_udp_size(m) << "\t"
		   << ends_extended_rcode << "\t"
		   << edns_version << "\t" << ldns_pkt_edns_z(m) << "\t"
		   << payload_len;
		if (output_ip_ttl) {
			ss << "\t" << ip_ttl_str;
		}
	};
	string null_prefix_str = "";
	string output_prefix_str = ss.str();
	
	if (PT) {
		ss << endl; 
		PT->put(ss.str());
	};

	if (QT)
		do_section(m, LDNS_SECTION_QUESTION, msgCount, dnsid_str, QT,
		    ((anon_flag & ANON_NAME) != ANON_NONE),
		    enc_key, key_len, enc_iv, iv_len, pbits,
		    (report_qry == UNIFIED_REPORT ? output_prefix_str : null_prefix_str));
	
	if (RT) {
		string rsp_prefix_str = (report_rsp == UNIFIED_REPORT ? output_prefix_str : null_prefix_str);
		do_section(m, LDNS_SECTION_ANSWER, msgCount, dnsid_str, RT,
		    ((anon_flag & ANON_NAME) != ANON_NONE),
		    enc_key, key_len, enc_iv, iv_len, pbits, rsp_prefix_str);
			
		do_section(m, LDNS_SECTION_AUTHORITY, msgCount, dnsid_str, RT,
		    ((anon_flag & ANON_NAME) != ANON_NONE),
		    enc_key, key_len, enc_iv, iv_len, pbits, rsp_prefix_str);
			
		do_section(m, LDNS_SECTION_ADDITIONAL, msgCount, dnsid_str, RT,
		    ((anon_flag & ANON_NAME) != ANON_NONE),
		    enc_key, key_len, enc_iv, iv_len, pbits, rsp_prefix_str);
	}

	ldns_pkt_free(m);
}

/*
	deal with each packet

	either process them (UDP) or pass to TCP reassembly

*/
void
per_packet(libtrace_packet_t *packet, string &outputPath, int pbits)
{
	uint32_t remaining = 0;
	uint8_t proto = 0;		
	void *transport_header = trace_get_transport(packet, &proto, &remaining);	

	if (!transport_header || !(proto == IPPROTO_UDP || proto == IPPROTO_TCP)) {
		// cerr << "non-tcp / upd " << proto << endl;
		return;
	};

	uint16_t src_port = 0;			//get port numbers
	uint16_t dst_port = 0;
	dst_port = trace_get_destination_port(packet);
	src_port = trace_get_source_port(packet); 
	
	if (!(dst_port == 53 || src_port == 53)) {	//not DNS packet
		return;
	}
		
	int dns_msg_len = 0;
	libtrace_udp_t *udp_header = NULL;
	libtrace_tcp_t *tcp_header = NULL;
	int tcp_payload_size = -1;
	u_char * payload = NULL;
	string err_state = "";
	
	size_t payload_length_without_padding = trace_get_payload_length(packet);
	if (proto == IPPROTO_UDP) {	
		udp_header = (libtrace_udp_t *)transport_header;
		dns_msg_len = ntohs(udp_header->len);
		/* xxx: should compare length vs. capture_length to check for truncation */
		payload = (u_char *)trace_get_payload_from_udp(udp_header, &remaining);
		per_message(packet, payload, dns_msg_len, outputPath, pbits);
	} else if (proto == IPPROTO_TCP) {	
		tcp_header = (libtrace_tcp_t *)transport_header;
		uint32_t post_tcp_remaining;
		payload = (u_char *)trace_get_payload_from_tcp(tcp_header, &post_tcp_remaining);
		/* xxx: payload could be null if incomplete header */
		int cl = trace_get_capture_length(packet);
		int wl = trace_get_wire_length(packet);
		if (!(cl == wl || cl == wl-4)) {
			/* capture seems always off by 4 bytes */
			err_state = "SNAPLEN_TRUNCATE";
		} else {
			// *remaining* should not be used as tcp length here because it includes ethernet padding
			// http://www.wand.net.nz/pipermail/libtrace-users/2013-February/000038.html
			// dnsanon_pcap_tcp_handler(packet, tcp_header, remaining, payload, outputPath, pbits);

			// add back offset for tcp
			size_t tcp_length = payload_length_without_padding + (4 * tcp_header->doff);
			dnsanon_pcap_tcp_handler(packet, tcp_header, tcp_length, payload, outputPath, pbits);
			/* packet parsing will happen in our callback later */
			return;
		};
	} else {
		die("per_packet: unknown proto\n");
	};

	if ((tcp_payload_size == 0 || remaining == 0) && report_extra) {
#if 0
		stringstream ss;
		ss.str("");
		ss << timestamp_str << "\t" << anon_src_ip_str << "\t" << src_port << "\t"
		   << anon_dst_ip_str << "\t" << dst_port << "\t" << remaining;
		if (err_state == "")
			err_state = PAYLOADZERO;
		if (proto == IPPROTO_TCP) {
			out_extra_info(ss.str(), err_state, tcp_header, true, TCPET);
		} else if (proto == IPPROTO_UDP) {
			out_extra_info(ss.str(), err_state, NULL, false, UDPET);
		}
#endif /* 0 */
		return;
	}
		
}

/*
	main function for parsing
*/
void
parse (string inputPath, string outputPath, string keyFile, int pbits, string out_pfx, vector<string> cmp)
{
#ifdef DEBUG
	cout<<"start parsing..."<<endl;
#endif
	timerclear(&check_tcp_ts);
	if (output_ip_ttl) {
		PT_sch.push_back("ipttl");
	}

	//bool anon = (anon_flag.size() != 0);
	//bool anon_response = (anon_flag == "wr");
	
	//check the valid path, input should be a file, output should be directory
	//if (outputPath.length() > 0) {
	//	check_path((char *)outputPath.c_str(), true);
	//}
	//check_path((char *)inputPath.c_str(), false);
	output_prefix = out_pfx;
	compressor = cmp;
	
	libtrace_t *trace		= NULL;		//trace
	libtrace_packet_t *packet 	= NULL;		//packet
	packet = trace_create_packet();			//Creating and initialising a packet structure
    	
        //everything is ok and continue
	if (packet == NULL) {
		perror("Creating libtrace packet");
		libtrace_cleanup(trace, packet);
		exit(-1);
	}

	//open trace
	trace = trace_create(inputPath.c_str());
	if (trace_is_err(trace)) {
		trace_perror(trace,"Opening trace file");
		libtrace_cleanup(trace, packet);
		exit(-1);
	}

	if (trace_start(trace) == -1) {
		trace_perror(trace,"Starting trace");
		libtrace_cleanup(trace, packet);
		exit(-1);
	}
	
	if (anon_flag!=ANON_NONE) {
		get_enc_key (keyFile);
	}
	
	//set the first pkt flag to get the time stamp as file name
	//bool firstpkt = true;
	
	//signal
	signal (SIGQUIT, sig_handler);
	signal (SIGINT, sig_handler);
	
	//loop all the packets
	while (trace_read_packet(trace, packet) > 0) {
		per_packet(packet, outputPath, pbits);
	}

	if (trace_is_err(trace)) {
		trace_perror(trace,"Reading packets");
		libtrace_cleanup(trace, packet);
		exit(-1);
	}
	
	clean_all_table();			//close table
	
	libtrace_cleanup(trace, packet);	//libtrace clean up
	
	if (enc_key != NULL) 
		delete[] enc_key;
	
	if (enc_iv != NULL) 
		delete[] enc_iv;

	if (verbose_log) {
		cout << "Parsing ends... " << endl 
		     << "Total "<<msgCount << " DNS packets..."<<endl;
	}
}

