/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "table.hh"
#include <iostream>
#include <sstream>
#include <cstring>
#include <err.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <algorithm>
using namespace std;

#define PIPEBUFSZ (8*1024*1024)

void check_empty(vector<string> s, const char *info)
{
	if (s.size() == 0)
		errx(1, "[error] %s is invalid", info);
}

void check_empty(string s, const char *info)
{
	if (s.length() == 0)
		errx(1, "[error] %s is invalid", info);
}

Table::Table (vector<string> sch, string fn, vector<string> cmp) 
{
	schema = sch;
	compressor = cmp;
	file_name = fn;
	is_stdout = (fn == "-" || fn[0] == '-');

	check_empty(schema, "table schema");
	check_empty(compressor, "compressor");
	check_empty(file_name, "file name");

	if (!is_stdout) {
		f = fopen (file_name.c_str(), "w");
		if (!f)
			err(1, "[error] Unable to create file[%s]", file_name.c_str());
		
		fd = fileno(f);
		if (fd == -1)
			err(1, "[error] fail to get fd for file[%s]", file_name.c_str());
		
		if (compressor[0] != "none")
			map_pipe();
		
		stringstream ss; //write the first line
		ss << "#fsdb -F t ";
		for (string s : schema)
			ss << s << " ";
		ss << endl;
		put(ss.str());
	} else {
		cout << "#fsdb -F t ";
		for (string s : schema)
			cout << s << " ";
		cout << endl;
	}
}

Table::~Table()
{
	clean_up();
}

void Table::clean_up()
{
	if (fd != -1)
		close(fd);
	if (f)
		fclose(f);
	f = NULL;
	fd = -1;
}

void Table::map_pipe()
{
	int pipefd[2];
	int i = 0, maxfd = getdtablesize();
	pid_t compressor_pid;
	if (0 != socketpair(AF_UNIX, SOCK_STREAM, 0, pipefd))
		err(1, "[error] fail to create pipe");
#ifdef HAVE_SETSOCKOPT
	int bufsz;
	bufsz = PIPEBUFSZ;
	if (0 != setsockopt(pipefd[0], SOL_SOCKET, SO_RCVBUF, &bufsz, sizeof(bufsz)))
		warn("setsockopt SO_RCVBUF failed\n");
	bufsz = PIPEBUFSZ;
	if (0 != setsockopt(pipefd[1], SOL_SOCKET, SO_SNDBUF, &bufsz, sizeof(bufsz)))
		warn("setsockopt SO_SNDBUF failed\n");
#endif
	compressor_pid =
# ifdef HAVE_FORK
	    fork()
# else
	    vfork()
# endif
	    ;
	    
	if (compressor_pid == -1)
		err(1, "[error] fail to fork compressor");
	if (compressor_pid == 0) { //child
#ifdef HAVE_SETPGID
		setpgid(0,0);
#endif
		if (-1 == dup2(pipefd[0], STDIN_FILENO))
			err(1, "[error] fail to duplicate compressor stdin");
		close(pipefd[0]);
		if (-1 == dup2(fd, STDOUT_FILENO))
			err(1, "[error] fail to duplicate compressor stdout");
		for (i = 0; i<maxfd; ++i) {
			if (i != STDIN_FILENO && i != STDOUT_FILENO && i != STDERR_FILENO)
				close(i);
		}

		vector<char *> cs;
		string stdout_option = "-c";
		bool found = false;
		for (string s : compressor) {
			cs.push_back(const_cast<char *>(s.c_str()));
			if (s == stdout_option) found = true;
		}
		if (!found)
			cs.push_back(const_cast<char *>(stdout_option.c_str()));
		cs.push_back(NULL);

		execvp(cs[0], &cs[0]);
		err(1, "[error] failed to execute pipe command %s", cs[0]);
	} else { //parent
		close(pipefd[0]);
		dup2(pipefd[1], fd);
		close(pipefd[1]);
	}
}

void Table::put(string s)
{
	if (is_stdout) {
		cout << s;
	} else {
		const char *cs = s.c_str();
		int len = strlen(cs);
		if (write(fd, (char *)cs, len) == -1)
			err(1, "[error] fail to write file");
	}
}

void Table::final(string cmd) 
{
	put(cmd);
	clean_up();
}
